/*	
 *	clipacm.c
 *
 *	USB CDC ACM (Communication Device Class Abstract Control Model)
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2007 Naoyuki Sawa
 *
 *	* Tue Oct 30 17:22:39 JST 2007 Naoyuki Sawa
 *	- 1st [XB
 *	- clipuioW[ɂč쐬܂B
 *	  ̓eclipuioƊSɓȂ̂ŁAȂ܂B
 *	  clipuio.c̃RgQƂĂB
 *	- clipuioƂ̈Ⴂ́Aȉ̓_łB
 *	  clipuio̓x_ƎNX̃foCXłApcecom.sysɂāAVA|[gƔFĂ܂B
 *	  clipacmCDC ACMNX̃foCXłAWindowsWusbser.sysɂāAVA|[gƔF܂B
 *	  ACXg[̂݁AƎcdcacm.infKvłB(usbser.sys𗘗pfoCX̋ʎdlłB)
 *	* Wed Oct 31 22:31:30 JST 2007 Naoyuki Sawa
 *	- SERIAL_STATE NotificationɑΉ܂B
 *	* Sun Nov 04 14:04:02 JST 2007 Naoyuki Sawa
 *	- (DTR)(DSR)łȂA(DSR)(CD)Ƀ[vobN悤ύX܂B
 */
#include "clip.h"

/****************************************************************************
 *	fBXNv^`
 ****************************************************************************/

static const DEVICE_DESCRIPTOR device_desc = {
	bLength:		sizeof(DEVICE_DESCRIPTOR),
	bDescriptorType:	DEVICE_DESCRIPTOR_TYPE,
	bcdUSB:			0x0110,
	bDeviceClass:		0x02,			/* Communication Device Class */
	bMaxPacketSize0:	EP0_PACKET_SIZE,
	idVendor:		PCE_VENDOR_ID,
	idProduct:		CDCACM_PRODUCT_ID,	/* d֎~!! */
	bcdDevice:		0x0001,
	iManufacture:		1,
	iProduct:		2,
	bNumConfigurations:	1,
};

/*--------------------------------------------------------------------------*/

static const CONFIGURATION_DESCRIPTOR config_desc = {
	bLength:		sizeof(CONFIGURATION_DESCRIPTOR),
	bDescriptorType:	CONFIGURATION_DESCRIPTOR_TYPE,
	wTotalLength:		sizeof(CONFIGURATION_DESCRIPTOR) +	/* config_desc */
				sizeof(INTERFACE_DESCRIPTOR) +		/* interface_desc_0 */
				(5+4+5+5) +				/* cdc_class_specific_desc_0 */
				sizeof(ENDPOINT_DESCRIPTOR) +		/* endpoint_desc_ep1in */
				sizeof(INTERFACE_DESCRIPTOR) +		/* interface_desc_1 */
				sizeof(ENDPOINT_DESCRIPTOR) +		/* endpoint_desc_ep2out */
				sizeof(ENDPOINT_DESCRIPTOR),		/* endpoint_desc_ep2in */
	bNumInterfaces:		2,
	bConfigurationValue:	1,
	bmAttributes:		0x80,		/* Bus-powered */
	MaxPower:		100 / 2,	/* 100[mA] () */
};
static const INTERFACE_DESCRIPTOR interface_desc_0 = {
	bLength:		sizeof(INTERFACE_DESCRIPTOR),
	bDescriptorType:	INTERFACE_DESCRIPTOR_TYPE,
	bInterfaceNumber:	0,
	bNumEndpoints:		1,
	bInterfaceClass:	0x02,	/* Communication Interface Class */
	bInterfaceSubClass:	0x02,	/* Abstract Control Model */
	bInterfaceProtocol:	0x01,	/* V.25ter */
};
static const unsigned char cdc_class_specific_desc_0[5+4+5+5] = {
	5,		/* bFunctionLength */
	0x24,		/* bDescriptorType = CS_INTERFACE */
	0x00,		/* bDescriptorSubtype = Header Functional Descriptor */
	0x10, 0x01,	/* bcdCDC */
	//
	4,		/* bFunctionLength */
	0x24,		/* bDescriptorType = CS_INTERFACE */
	0x02,		/* bDescriptorSubtype = Abstract Control Management Functional Descriptor */
	0x02,		/* bmCapabilities = SET_LINE_CODING,GET_LINE_CODING,SET_CONTROL_LINE_STATE,SERIAL_STATE */
	//
	5,		/* bFunctionLength */
	0x24,		/* bDescriptorType = CS_INTERFACE */
	0x06,		/* bDescriptorSubtype = Union Functional descriptor */
	0,		/* bMasterInterface */
	1,		/* bSlaveInterface0 */
	//
	5,		/* bFunctionLength */
	0x24,		/* bDescriptorType = CS_INTERFACE */
	0x01,		/* bDescriptorSubtype = Call Management Functional Descriptor */
	0x00,		/* bmCapabilities */
	1,		/* bDataInterface */
};
static const ENDPOINT_DESCRIPTOR endpoint_desc_ep1in = {
	bLength:		sizeof(ENDPOINT_DESCRIPTOR),
	bDescriptorType:	ENDPOINT_DESCRIPTOR_TYPE,
	bEndpointAddress:	0x81,				/* EP1IN */
	bAttributes:		3,				/* Interrupt */
	wMaxPacketSize:		EP1_PACKET_SIZE,
	bInterval:		2,				/* 2[ms] */
};
static const INTERFACE_DESCRIPTOR interface_desc_1 = {
	bLength:		sizeof(INTERFACE_DESCRIPTOR),
	bDescriptorType:	INTERFACE_DESCRIPTOR_TYPE,
	bInterfaceNumber:	1,
	bNumEndpoints:		2,
	bInterfaceClass:	0x0A,				/* Data Interface Class */
};
static const ENDPOINT_DESCRIPTOR endpoint_desc_ep2out = {
	bLength:		sizeof(ENDPOINT_DESCRIPTOR),
	bDescriptorType:	ENDPOINT_DESCRIPTOR_TYPE,
	bEndpointAddress:	0x02,				/* EP2OUT */
	bAttributes:		2,				/* Bulk */
	wMaxPacketSize:		EP2_NONISO_PACKET_SIZE,
};
static const ENDPOINT_DESCRIPTOR endpoint_desc_ep2in = {
	bLength:		sizeof(ENDPOINT_DESCRIPTOR),
	bDescriptorType:	ENDPOINT_DESCRIPTOR_TYPE,
	bEndpointAddress:	0x82,				/* EP2IN */
	bAttributes:		2,				/* Bulk */
	wMaxPacketSize:		EP2_NONISO_PACKET_SIZE,
};

/*--------------------------------------------------------------------------*/

static const char* const string_desc[] = {
//	   ++-- 04 06 08 0a 0c 0e 10 12 14 16 18 1a 1c 1e 20 22 24 26 28 2a 2c 2e 30
//	   ||
//	"\x??\3_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0"
	"\x04\3\x09\x04",						// 0: wLANGID = 0x0409 English (United States)
	"\x16\3P\0i\0e\0c\0e\0 \0L\0a\0b\0.\0",				// 1: iManufacture   = "Piece Lab."
	"\x22\3P\0/\0E\0C\0E\0 \0U\0S\0B\0 \0d\0e\0v\0i\0c\0e\0",	// 2: iProduct       = "P/ECE USB device"
};

/****************************************************************************
 *	
 ****************************************************************************/

typedef struct _CDCACM {
	unsigned char* rxbuf/*[rxcap]*/;
	int rxcap;
	int rxpos;
	int rxlen;
	int rxlog;
	int rxerr;
	void (*rxdone)();
	//
	unsigned char* txbuf/*[txcap]*/;
	int txcap;
	int txpos;
	int txlen;
	int txlog;
	int txing;
	void (*txdone)();
} CDCACM;
static CDCACM cdcacm;

static void get_descriptor(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
static void set_configuration(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
static void set_interface(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
static void class_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
static void ep2_rxdone(const unsigned char dat[/*len*/], int len);
static void ep2_txdone();
static void ep2_txsub();

static const USBINF usb_inf = {
	/* USBC mode */
	mode:			D12_ENDP_NONISO,
	/* USBC raw event */
	/* Communication Class Interface */
/*{{gp*/
//	ep1_txdone:		ep1_txdone,
/*}}gp*/
	/* Data Class Interface */
	ep2_rxdone:		ep2_rxdone,
	ep2_txdone:		ep2_txdone,
	/* Device request */
	get_descriptor:		get_descriptor,
	set_configuration:	set_configuration,
	set_interface:		set_interface,
	/* Class request */
	class_request:		class_request,
};

/* * Wed Oct 31 22:31:30 JST 2007 Naoyuki Sawa
 * - xłSERIAL_STATEʒm(CTS)=1ɂȂAx(CTS)=0ɖ߂܂B
 *   usbser.sys̎dl݂łB(Win2K SP4ɂĎ)
 * - wIndexɂ́ACommunication Class InterfaceA܂́AData Class InterfaceA
 *   ǂInterfaceԍw肷ׂȂ̂słB
 *   Win2K SP4ŎƂAǂł삵܂B
 *   wIndex=0(Communication Class Interface)ƂĂƂɂ܂B
 * * Sun Nov 04 14:43:12 JST 2007 Naoyuki Sawa
 * - (CTS)=0ɖ߂Ȃ̗RɂāA[L܂B
 *	u˓cF̎ݔiΔ(http://www.lbm.go.jp/toda/)v
 *	uRS-232CP[ũNXdlɊւl@(http://www.lbm.go.jp/toda/comp/rscross.html)v
 *   ukeep/RS-232CP[ũNXdlɊւl@.7zvɕۑĂ̂ŁAQƂĂB
 */
typedef struct _SerialStateNotification {
	unsigned char bmRequestType;	/* +0,1 = 0xA1 (In,ClassRequest,1)*/
	unsigned char bNotification;	/* +1,1 = 0x20 (SERIAL_STATE) */
	unsigned short wValue;		/* +2,2 = 0 (Œ) */
	unsigned short wIndex;		/* +4,2 = 0 (Interface) */
	unsigned short wLength;		/* +6,2 = 2 (DataLength) */
	/*{{UART State Bitmap*/		/* +8,2 */
	unsigned short bRxCarrier  : 1;	/*      = (CD) */
	unsigned short bTxCarrier  : 1;	/*      = (DSR) */
	unsigned short bBreak      : 1;
	unsigned short bRingSignal : 1;	/*      = (RI) */
	unsigned short bFraming    : 1;
	unsigned short bParity     : 1;
	unsigned short bOverRun    : 1;
	unsigned short             : 9;
	/*}}UART State Bitmap*/
} SerialStateNotification;		/* =10 */
static SerialStateNotification serial_state_notification = { 0xA1,0x20,0,0,2 };

/****************************************************************************
 *	
 ****************************************************************************/

static void
get_descriptor(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	static unsigned char tmpbuf[sizeof(CONFIGURATION_DESCRIPTOR) +	/* config_desc */
				    sizeof(INTERFACE_DESCRIPTOR) +	/* interface_desc_0 */
				    (5+4+5+5) +				/* cdc_class_specific_desc_0 */
				    sizeof(ENDPOINT_DESCRIPTOR) +	/* endpoint_desc_ep1in */
				    sizeof(INTERFACE_DESCRIPTOR) +	/* interface_desc_1 */
				    sizeof(ENDPOINT_DESCRIPTOR) +	/* endpoint_desc_ep2out */
				    sizeof(ENDPOINT_DESCRIPTOR)];	/* endpoint_desc_ep2in */
	//
	int type  = HIBYTE(req->wValue);
	int index = LOBYTE(req->wValue);

	switch(type) {
	case DEVICE_DESCRIPTOR_TYPE:
		usb_ep0_tx(&device_desc, sizeof(DEVICE_DESCRIPTOR));
		return;
	case CONFIGURATION_DESCRIPTOR_TYPE:
		switch(index) {
		case 0:
			len = 0;
			memcpy(&tmpbuf[len], &config_desc              , sizeof(CONFIGURATION_DESCRIPTOR)); len += sizeof(CONFIGURATION_DESCRIPTOR);
			memcpy(&tmpbuf[len], &interface_desc_0         , sizeof(INTERFACE_DESCRIPTOR)    ); len += sizeof(INTERFACE_DESCRIPTOR)    ;
			memcpy(&tmpbuf[len], &cdc_class_specific_desc_0, (5+4+5+5)                       ); len += (5+4+5+5)                       ;
			memcpy(&tmpbuf[len], &endpoint_desc_ep1in      , sizeof(ENDPOINT_DESCRIPTOR)     ); len += sizeof(ENDPOINT_DESCRIPTOR)     ;
			memcpy(&tmpbuf[len], &interface_desc_1         , sizeof(INTERFACE_DESCRIPTOR)    ); len += sizeof(INTERFACE_DESCRIPTOR)    ;
			memcpy(&tmpbuf[len], &endpoint_desc_ep2out     , sizeof(ENDPOINT_DESCRIPTOR)     ); len += sizeof(ENDPOINT_DESCRIPTOR)     ;
			memcpy(&tmpbuf[len], &endpoint_desc_ep2in      , sizeof(ENDPOINT_DESCRIPTOR)     ); len += sizeof(ENDPOINT_DESCRIPTOR)     ;
			usb_ep0_tx(tmpbuf, len);
			return;
		}
		break;
	case STRING_DESCRIPTOR_TYPE:
		if(index < ARRAY_SIZE(string_desc)) {
			usb_ep0_tx(string_desc[index], string_desc[index][0]);
			return;
		}
		break;
	}
	usb_ep0_stall(); /* Ή */
}

static void
set_configuration(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	D12_SetEndpointEnable(req->wValue == 1);
	usb_ep0_tx(NULL, 0); /* ACK */
}

static void
set_interface(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	usb_ep0_tx(NULL, 0); /* ACK */
}

static void
class_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	static unsigned char line_coding_structure[7];

	switch(req->bRequest) {
/*{{Ή*/
//	/* SEND_ENCAPSULATED_COMMAND (Required) */
//	case 0x00:
//	/* GET_ENCAPSULATED_RESPONSE (Required) */
//	case 0x01:
/*}}Ή*/
	/* SET_LINE_CODING (Optional) */
	case 0x20:
		memcpy(line_coding_structure, dat, 7);
		usb_ep0_tx(NULL, 0); /* ACK */
		return;
	/* GET_LINE_CODING (Optional) */
	case 0x21:
		usb_ep0_tx(line_coding_structure, 7);
		return;
	/* SET_CONTROL_LINE_STATE (Optional) */
	case 0x22:
		usb_ep0_tx(NULL, 0); /* ACK */
		/* * Wed Oct 31 22:31:30 JST 2007 Naoyuki Sawa
		 * - (DTR)(DSR)Ƀ[vobN邱Ƃɂ܂B
		 * - AInterrupt|[OԊuZԊuSET_CONTROL_LINE_STATEāA
		 *   O"D12_WriteEndpoint(EP1IN,`)"ɂȂĂA肠܂B
		 *   SERIAL_STATE{bRxCarrier,bTxCarrier}́AŐVԂ݂̂ʒmΗǂłB
		 *   A{bBreak,bRingSignal,bFraming,bParity,bOverRun}̓gKʒmȂ̂ŁA
		 *   Ɩ肪܂A2007/10/31݂͎gĂȂ̂ővłB
		 * - SERIAL_STATEʒm邱ƂɂāAWindows͎I(CTS)=1Ɣf܂B
		 *   ڂ́ASerialStateNotification\̂̏ɏRgQƂĂB
		 */
		//serial_state_notification.bTxCarrier = (req->wValue & 1);
		//* Sun Nov 04 14:04:02 JST 2007 Naoyuki Sawa
		//- (DTR)(DSR)łȂA(DSR)(CD)Ƀ[vobN悤ύX܂B
		//  ʓIȃNXP[uڑ͈ȉ̂悤Ȃ̂ŁAɏ]߂łB
		//
		//	 DTE(p\R) 			 DTE(p\R) 
		//	-----------------			-----------------
		//	M sԍ			M sԍ
		//	   GND       1   -----------------------   GND       1   
		//	   TxD       2   -----------+   +-------   TxD       2   
		//	                        +---|---+                        
		//	   RxD       3   <------+   +---------->   RxD       3   
		//	   RTS       4   -----------+   +-------   RTS       4   
		//	                        +---|---+                        
		//	   CTS       5   <------+   +---------->   CTS       5   
		//	   DSR       6   <--+               +-->   DSR       6   
		//	   GND       7   ---|---------------|---   GND       7   
		//	   CD        8   <--+-------+   +---+-->   CD        8   
		//	   TxC2     15          +---|---+          TxC2     15   
		//	   RxC      17          |   |              RxC      17   
		//	   DTR      20   -------+   +-----------   DTR      20   
		//	   RI       22   			   RI       22   
		//	   TxC1     24   			   TxC1     24   
		//
		//	wgI/OC^tF[XbuxFrv,ĉj
		//	p.159}4-14:RS-232C𗘗pDTEm̐ڑ(NXP[ugڑ)
		//- ɂ́A(CD)ĎĂAvP[V͂قƂǖƎv̂ŁA
		//  ̕ύX͂܂Ӗm܂񂪁AÔ߂ɕύXĂƂɂ܂B
		//
		if(req->wValue & 1) { /* (DTR) */
			serial_state_notification.bRxCarrier = 1; /* (CD) */
			serial_state_notification.bTxCarrier = 1; /* (DSR) */
		} else {
			serial_state_notification.bRxCarrier = 0; /* (CD) */
			serial_state_notification.bTxCarrier = 0; /* (DSR) */
		}
		D12_WriteEndpoint(EP1IN, &serial_state_notification, 10);
		return;
	}
	usb_ep0_stall(); /* Ή */
}

/****************************************************************************
 *	
 ****************************************************************************/

static void
ep2_rxdone(const unsigned char dat[/*len*/], int len)
{
	int pos;
	int i;

	if(!len) {
		return;
	}
	if(len > (cdcacm.rxcap - cdcacm.rxlen)) {
		cdcacm.rxerr += len - (cdcacm.rxcap - cdcacm.rxlen);
				len = (cdcacm.rxcap - cdcacm.rxlen);
	}
	pos = cdcacm.rxpos + cdcacm.rxlen;
	if(pos >= cdcacm.rxcap) {
		pos -= cdcacm.rxcap;
	}
	for(i = 0; i < len; i++) {
		cdcacm.rxbuf[pos++] = *dat++;
		if(pos == cdcacm.rxcap) {
			pos = 0;
		}
	}
	cdcacm.rxlen += len;
	if(cdcacm.rxdone) {
		cdcacm.rxdone();
	}
}

static void
ep2_txdone()
{
	if(cdcacm.txlen) {
		ep2_txsub();
	} else {
		cdcacm.txing = 0;
		if(cdcacm.txdone) {
			cdcacm.txdone();
		}
	}
}

static void
ep2_txsub()
{
	unsigned char dat[EP2_NONISO_PACKET_SIZE];
	int len;
	int i;

	len = EP2_NONISO_PACKET_SIZE;
	if(len > cdcacm.txlen) {
		len = cdcacm.txlen;
	}
	for(i = 0; i < len; i++) {
		dat[i] = cdcacm.txbuf[cdcacm.txpos++];
		if(cdcacm.txpos == cdcacm.txcap) {
			cdcacm.txpos = 0;
		}
	}
	cdcacm.txlen -= len;
	D12_WriteEndpoint(EP2IN, dat, len);
}

/****************************************************************************
 *	
 ****************************************************************************/

void
cdcacm_start(int rxcap, int txcap, void (*rxdone)(), void (*txdone)())
{
	cdcacm_stop();
	memset(&cdcacm, 0, sizeof cdcacm);
	if(rxcap < EP2_NONISO_PACKET_SIZE) {
		rxcap = EP2_NONISO_PACKET_SIZE;
	}
	cdcacm.rxcap = rxcap;
	cdcacm.rxbuf = malloc(rxcap);
	if(!cdcacm.rxbuf) {
		DIE();
	}
	cdcacm.rxdone = rxdone;
	if(txcap < EP2_NONISO_PACKET_SIZE) {
		txcap = EP2_NONISO_PACKET_SIZE;
	}
	cdcacm.txcap = txcap;
	cdcacm.txbuf = malloc(txcap);
	if(!cdcacm.txbuf) {
		DIE();
	}
	cdcacm.txdone = txdone;
	usb_start(&usb_inf);
}

void
cdcacm_stop()
{
	usb_stop();
	free(cdcacm.rxbuf);
	free(cdcacm.txbuf);
	memset(&cdcacm, 0, sizeof cdcacm);
}

void
cdcacm_reset()
{
ENTER_CS;
	cdcacm.rxpos = 0;
	cdcacm.rxlen = 0;
	cdcacm.rxlog = 0;
	cdcacm.rxerr = 0;
	cdcacm.txpos = 0;
	cdcacm.txlen = 0;
	cdcacm.txlog = 0;
LEAVE_CS;
}

int
cdcacm_recv(void* _dat, int len)
{
	unsigned char* dat = _dat;
	int i;

ENTER_CS;
	if(!cdcacm.rxerr) {
		if(len > cdcacm.rxlen) {
			len = cdcacm.rxlen;
		}
		for(i = 0; i < len; i++) {
			*dat++ = cdcacm.rxbuf[cdcacm.rxpos++];
			if(cdcacm.rxpos == cdcacm.rxcap) {
				cdcacm.rxpos = 0;
			}
		}
		cdcacm.rxlen -= len;
		cdcacm.rxlog += len;
	} else {
		len = -cdcacm.rxerr;
	}
LEAVE_CS;
	return len;
}

int
cdcacm_send(const void* _dat, int len)
{
	const unsigned char* dat = _dat;
	int pos;
	int i;

	if(len < 0) {
		len = strlen((char*)dat);
	}
ENTER_CS;
	if(len > (cdcacm.txcap - cdcacm.txlen)) {
		len = (cdcacm.txcap - cdcacm.txlen);
	}
	if(len) {
		pos = cdcacm.txpos + cdcacm.txlen;
		if(pos >= cdcacm.txcap) {
			pos -= cdcacm.txcap;
		}
		for(i = 0; i < len; i++) {
			cdcacm.txbuf[pos++] = *dat++;
			if(pos == cdcacm.txcap) {
				pos = 0;
			}
		}
		cdcacm.txlen += len;
		cdcacm.txlog += len;
		if(!cdcacm.txing) {
			cdcacm.txing = 1;
			ep2_txsub();
		}
	}
LEAVE_CS;
	return len;
}

void
cdcacm_stat(CDCACMSTAT* stat)
{
ENTER_CS;
	stat->rxlen = cdcacm.rxlen;
	stat->rxlog = cdcacm.rxlog;
	stat->rxerr = cdcacm.rxerr;
	stat->txlen = cdcacm.txlen;
	stat->txlog = cdcacm.txlog;
LEAVE_CS;
}
